//-----------------------------------------------------------------------------
// File: Fire.PAS
//
// Desc:
//     A simple FIRE demo which demonstrating direct access to
//     video memory With pseudo MEM[] arrays
//
// Copyright (c) 2000 TMT Development Corp. All rights reserved.
// Ported by Vadim Bodrov, TMT Development Corp.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------
uses Windows, Messages, DXLite;

{$r fire.res}

const
    NAME  = 'FIRE';
    TITLE = 'Direct Draw File Example';

var
    Wnd: HWnd;
    g_bActive: Boolean;
    g_bIsInitialized: Boolean := FALSE;
    g_pDD: PPDirectDraw := nil;
    g_pDDSPrimary: PPDirectDrawSurface := nil;
    g_pDDPal: PPDirectDrawPalette := nil;

//-----------------------------------------------------------------------------
// Name: ReleaseAllObjects
// Desc: Finished with all objects we use; release them
//-----------------------------------------------------------------------------
procedure ReleaseAllObjects;
begin
    if g_pDD <> nil then
    begin
        if g_pDDSPrimary <> nil then
        begin
            g_pDDSPrimary^^.Release^(g_pDDSPrimary);
            g_pDDSPrimary := nil;
        end;
        if g_pDDPal <> nil then
        begin
            g_pDDPal^^.Release^(g_pDDPal);
            g_pDDPal := nil;
        end;
        g_pDD^^.Release^(g_pDD);
        g_pDD := nil;
    end;
end;

//-----------------------------------------------------------------------------
// Name: UpdateFrame
// Desc: Draws the fire
//-----------------------------------------------------------------------------
procedure UpdateFrame;
var
    SurfPtr: DWORD;
    desc: TDDSurfaceDesc;
    c, x, y: DWORD;
begin
    ZeroMemory(@desc, SizeOf(desc));
    desc.dwSize := SizeOf(desc);
    if g_pDDSPrimary^^.Lock^(g_pDDSPrimary, PRect(nil), desc, DDLOCK_SURFACEMEMORYPTR, 0) = DD_OK then
    begin
        SurfPtr := DWORD(desc.lpSurface);
        x := 0;
        repeat
            y := 80;
            repeat
                c := (mem[SurfPtr + y * 320 + x] +
                      mem[SurfPtr + y * 320 + x + 2]+
                      mem[SurfPtr + y * 320 + x - 2]+
                      mem[SurfPtr + (y+2) * 320 + x + 2]) div 4;
            if c > 1 then dec(c, 1);
            c := (c shl 8) + c;
            memw[SurfPtr + (y-2) * 320 + x] := c;
            memw[SurfPtr + (y-1) * 320 + x] := c;
            y +:= 2;
        until y >= 198;
        Dec(y,2);
        mem[SurfPtr + y * 320 + x] := random(12) * 10 + 8;
        X +:= 2;
        until x >= 320;
        g_pDDSPrimary^^.Unlock^(g_pDDSPrimary, nil);
    end;
    g_pDD^^.WaitForVerticalBlank^(g_pDD, DDWAITVB_BLOCKEND, 0)
end;

//-----------------------------------------------------------------------------
// Name: WindowProc
// Desc: The Main Window Procedure
//-----------------------------------------------------------------------------
function WindowProc conv arg_stdcall (Wnd: HWND; message: UINT; wParam: WPARAM; lParam: LPARAM): Longint;
begin
    case message of
      WM_SYSKEYUP:
          declare
          var
              Msg : TMsg;
          begin
              GetMessage(Msg, Wnd, 0, 0);
          end;

      WM_ACTIVATE:
          g_bActive := Boolean(wParam);

      WM_KEYDOWN:
          if wParam in [VK_ESCAPE, VK_F12] then
              PostMessage(Wnd, WM_CLOSE, 0, 0);

      WM_DESTROY:
          begin
              ReleaseAllObjects;
              PostQuitMessage(0);
          end;
    end;
    Result := DefWindowProc(Wnd, message, wParam, lParam);
end;


//-----------------------------------------------------------------------------
// Name: InitFail
// Desc: This function is called if an initialization function fails
//-----------------------------------------------------------------------------
function InitFail(Wnd: hWnd; hRet: HResult; szError: PChar): HResult;
begin
    ReleaseAllObjects;
    MessageBox(Wnd, szError, TITLE, MB_OK);
    DestroyWindow(Wnd);
    Result := hRet;
end;


//-----------------------------------------------------------------------------
// Name: InitApp
// Desc: Do work required for every instance of the application:
//          Create the window, initialize data
//-----------------------------------------------------------------------------
function InitApp(Instance: THandle; nCmdShow: Longint): Boolean;
var
    wc: TWndClass;
    ddsd: TDDSurfaceDesc;
    hRet: HResult;
    pe: array[0..255] of TPaletteEntry;
    i: DWORD;
begin
    wc.style := CS_HREDRAW or CS_VREDRAW;
    wc.lpfnWndProc := @WindowProc;
    wc.cbClsExtra := 0;
    wc.cbWndExtra := 0;
    wc.hInstance := Instance;
    wc.hIcon := LoadIcon(0, MAKEINTRESOURCE(IDI_APPLICATION));
    wc.hCursor := LoadCursor(0, IDC_ARROW);
    wc.hbrBackground := 0;
    wc.lpszMenuName := nil;
    wc.lpszClassName := NAME;
    RegisterClass(wc);

    Wnd := CreateWindowEx(0,
                          NAME,
                          TITLE,
                          WS_POPUP,
                          0,
                          0,
                          GetSystemMetrics(SM_CXSCREEN),
                          GetSystemMetrics(SM_CYSCREEN),
                          0,
                          0,
                          hInstance,
                          nil);

    if Wnd = 0 then
    begin
        Result :=  FALSE;
        exit;
    end;
    ShowWindow(Wnd, nCmdShow);
    UpdateWindow(Wnd);
    ///////////////////////////////////////////////////////////////////////////
    // Create the main DirectDraw object
    ///////////////////////////////////////////////////////////////////////////
    hRet := DirectDrawCreate(nil, g_pDD, nil);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'DirectDrawCreateEx FAILED');
        Result := FALSE;
        exit;
    end;

    // Get exclusive mode
    hRet := g_pDD^^.SetCooperativeLevel^(g_pDD, Wnd, DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'SetCooperativeLevel FAILED');
        Result := FALSE;
        exit;
    end;

    // Set the video mode to 320x200x8
    hRet := g_pDD^^.SetDisplayMode^(g_pDD, 320, 200, 8);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'SetDisplayMode FAILED');
        Result := FALSE;
        exit;
    end;

    // Create the primary surface
    ZeroMemory(@ddsd, SizeOf(ddsd));
    ddsd.dwSize := SizeOf(ddsd);
    ddsd.dwFlags := DDSD_CAPS;
    ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE;
    hRet := g_pDD^^.CreateSurface^(g_pDD, ddsd, g_pDDSPrimary, nil);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'CreateSurface FAILED');
        Result := FALSE;
        exit;
    end;

   ZeroMemory(@pe, sizeof(pe));

    for i := 1 to 32 do
    begin
        pe[i].peRed := i * 8 - 1; pe[i].peGreen := 0; pe[i].peBlue := 0;
        pe[i+32].peRed := 255; pe[i+32].peGreen := i * 8 - 1; pe[i+32].peBlue := 0;
        pe[i+64].peRed := 255; pe[i+64].peGreen := 255; pe[i+64].peBlue := i * 8 - 1;
        pe[i+96].peRed := 255; pe[i+96].peGreen := 255; pe[i+96].peBlue := 255;
    end;
    FillChar(pe[1], SizeOf(pe[1]), 0);

    if g_pDDPal = nil then
    begin
      hRet := g_pDD^^.CreatePalette^(g_pDD, DDPCAPS_8BIT, @pe, g_pDDPal, nil);
      if hRet <> DD_OK then
      begin
          Result := FALSE;
          exit;
      end;
    end;

    g_pDDSPrimary^^.SetPalette^(g_pDDSPrimary, g_pDDPal);

    g_bIsInitialized := TRUE;
    Result := TRUE;
end;


//-----------------------------------------------------------------------------
// Name: WinMain
// Desc: Initialization, message loop
//-----------------------------------------------------------------------------
var
  msg: TMsg;
begin
    if not InitApp(hInstance, CmdShow) then
        halt(0);

    ShowCursor(FALSE);

    while (TRUE) do
    begin
        if PeekMessage(msg, 0, 0, 0, PM_NOREMOVE) then
        begin
            if not GetMessage(msg, 0, 0 ,0) then
            begin
                ShowCursor(TRUE);
                exit;
            end;
            TranslateMessage(msg);
            DispatchMessage(msg);
        end else
        if g_bActive then
        begin
            UpdateFrame;
        end else
            WaitMessage;
    end;
end.