//-----------------------------------------------------------------------------
// File: WormHole.PAS
//
// Desc:
//
// Copyright (c) 1995-1999 Microsoft Corporation. All rights reserved.
// Ported by Vadim Bodrov, TMT Development Corp.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------
uses Windows, Messages, DXLite;
{$r wormhole.res}

const
    IDI_MAIN_ICON = 101;

//-----------------------------------------------------------------------------
// Local definitions
//-----------------------------------------------------------------------------
const
     NAME   = 'WormHole';
     TITLE  = 'Direct Draw WormHole Example';

//-----------------------------------------------------------------------------
// Default settings
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// External functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// External data
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Internal functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Global data
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Local data
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Public functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Private functions
//-----------------------------------------------------------------------------

var
    Wnd: HWnd;
    pe: array[0..255] of TPaletteEntry;
    g_bActive: Boolean;
    g_bIsInitialized: Boolean := FALSE;

    g_pDD: PPDirectDraw := nil;
    g_pDDSPrimary: PPDirectDrawSurface := nil;
    g_pDDSOne: PPDirectDrawSurface := nil;
    g_pClipper: PPDirectDrawClipper := nil;
    g_pDDPal: PPDirectDrawPalette := nil;


function ReadBMPIntoSurfaces: Boolean;
var
    hRet: HResult;
    hBMP: HRSrc;
    Palette: array[0..255] of TRGBQuad;
    ddsd: TDDSurfaceDesc;
    pBits: PChar;
    pSrc: PChar;
    pBmp: Pointer;
    i: Longint;
begin

    hBMP := FindResource(0, 'wormhole_bmp', RT_BITMAP);
    if (hBMP = 0) then
    begin
        Result := FALSE;
        exit;
    end;

    pBMP := PCHar(LockResource(LoadResource(0, hBMP)));

    CopyMemory(@Palette, pBMP + SizeOf(TBITMAPINFOHEADER), SizeOf(Palette));

    FreeResource(hBMP);

    for i := 0 to 255 do
    begin
        pe[i].peRed := Palette[i].rgbRed;
        pe[i].peGreen := Palette[i].rgbGreen;
        pe[i].peBlue := Palette[i].rgbBlue;
    end;

    if g_pDDPal = nil then
    begin
      hRet := g_pDD^^.CreatePalette^(g_pDD, DDPCAPS_8BIT, @pe, g_pDDPal, nil);
      if hRet <> DD_OK then
      begin
          Result := FALSE;
          exit;
      end;
    end;

    g_pDDSPrimary^^.SetPalette^(g_pDDSPrimary, g_pDDPal);

    ddsd.dwSize := SizeOf(ddsd);
    hRet := g_pDDSOne^^.Lock^(g_pDDSOne, nil, ddsd, DDLOCK_WAIT, 0);
    if hRet <> DD_OK then
    begin
        Result := FALSE;
        exit;
    end;
    pBits := ddsd.lpSurface;
    pSrc := pBMP + SizeOf(TBITMAPINFOHEADER) + SizeOf(Palette) + (640*480);

    for i := 0 to 479 do
    begin
        dec(pSrc, 640);
        CopyMemory(pBits, pSrc, 640);
        inc(pBits, ddsd.lPitch);
    end;
    g_pDDSOne^^.Unlock^(g_pDDSOne, nil);
    Result := TRUE;
end;

//-----------------------------------------------------------------------------
// Name: ReleaseAllObjects()
// Desc: Finished with all objects we use; release them
//-----------------------------------------------------------------------------
procedure ReleaseAllObjects;
begin
    if g_pDD <> nil then
    begin
        if g_pDDSPrimary <> nil then
        begin
            g_pDDSPrimary^^.Release^(g_pDDSPrimary);
            g_pDDSPrimary := nil;
        end;
        if g_pDDSOne <> nil then
        begin
            g_pDDSOne^^.Release^(g_pDDSOne);
            g_pDDSOne := nil;
        end;
        if g_pDDPal <> nil then
        begin
            g_pDDPal^^.Release^(g_pDDPal);
            g_pDDPal := nil;
        end;
        g_pDD^^.Release^(g_pDD);
        g_pDD := nil;
    end;
end;

//-----------------------------------------------------------------------------
// Name: InitFail()
// Desc: This function is called if an initialization function fails
//-----------------------------------------------------------------------------
function InitFail(Wnd: hWnd; hRet: HResult; szError: PChar): HResult;
begin
    ReleaseAllObjects;
    MessageBox(Wnd, szError, TITLE, MB_OK);
    DestroyWindow(Wnd);
    Result := hRet;
end;

//-----------------------------------------------------------------------------
// Name: RestoreAll()
// Desc: Restore all lost objects
//-----------------------------------------------------------------------------
function RestoreAll: Boolean;
var
  bResult: Boolean;
begin
    bResult := (g_pDDSPrimary^^.Restore^(g_pDDSPrimary) = DD_OK) and
        (g_pDDSOne^^.Restore^(g_pDDSOne) = DD_OK);

    ReadBMPIntoSurfaces;

    Result := bResult;
end;

//-----------------------------------------------------------------------------
// Name: UpdateFrame()
// Desc: Decide what needs to be blitted next
//-----------------------------------------------------------------------------
procedure UpdateFrame;
var
    rcRect: TRect;
    destRect: TRect;
    hRet: HResult;
    pt: TPoint;
begin
   rcRect.left := 0;
    rcRect.top := 0;
    rcRect.right := 640;
    rcRect.bottom := 480;

    GetClientRect(Wnd, destRect);

    pt.x := 0;
    pt.y := 0;
    ClientToScreen(Wnd, pt);
    OffsetRect(destRect, pt.x, pt.y);

    while (TRUE) do
    begin
        hRet := g_pDDSPrimary^^.Blt^(g_pDDSPrimary, @destRect, g_pDDSOne, @rcRect, 0, nil);

        if hRet = DD_OK then
            break;

        if hRet = DDERR_SURFACELOST then
        begin
            if not RestoreAll then
                exit;
            continue;
        end;
        if hRet <> DDERR_WASSTILLDRAWING then
           exit;
    end;
end;

//-----------------------------------------------------------------------------
// Name: WindowProc()
// Desc: The Main Window Procedure
//-----------------------------------------------------------------------------
function WindowProc conv arg_stdcall (Wnd: HWND; message: UINT; wParam: WPARAM; lParam: LPARAM): Longint;
begin
    case message of
      WM_ACTIVATE:
          g_bActive := Boolean(wParam);

      WM_SETCURSOR:
          begin
            SetCursor(0);
            if g_bIsInitialized then
            begin
              UpdateFrame;
              g_pDDPal^^.GetEntries^(g_pDDPal, 0, 0, 256, PPaletteEntry(@pe));
            end;
          end;

      WM_KEYDOWN:
          if wParam in [VK_ESCAPE, VK_F12] then
              PostMessage(Wnd, WM_CLOSE, 0, 0);

      WM_DESTROY:
          begin
              ReleaseAllObjects;
              PostQuitMessage(0);
          end;
    end;
    Result := DefWindowProc(Wnd, message, wParam, lParam);
end;

//-----------------------------------------------------------------------------
// Name: InitApp()
// Desc: Do work required for every instance of the application:
//          Create the window, initialize data
//-----------------------------------------------------------------------------
function InitApp(Instance: THandle; nCmdShow: Longint): Boolean;
var
    wc: TWndClass;
    ddsd: TDDSurfaceDesc;
    hRet: HResult;
begin
    wc.style := CS_HREDRAW or CS_VREDRAW;
    wc.lpfnWndProc := @WindowProc;
    wc.cbClsExtra := 0;
    wc.cbWndExtra := 0;
    wc.hInstance := Instance;
    wc.hIcon := LoadIcon(Instance, MAKEINTRESOURCE(IDI_MAIN_ICON));
    wc.hCursor := LoadCursor(0, IDC_ARROW);
    wc.hbrBackground := 0;
    wc.lpszMenuName := NAME;
    wc.lpszClassName := NAME;
    RegisterClass(wc);

    Wnd := CreateWindowEx(0,
                          NAME,
                          TITLE,
                          WS_POPUP,
                          0,
                          0,
                          GetSystemMetrics(SM_CXSCREEN),
                          GetSystemMetrics(SM_CYSCREEN),
                          0,
                          0,
                          hInstance,
                          nil);

    if Wnd = 0 then
    begin
        Result :=  FALSE;
        exit;
    end;
    ShowWindow(Wnd, nCmdShow);
    UpdateWindow(Wnd);
    ///////////////////////////////////////////////////////////////////////////
    // Create the main DirectDraw object
    ///////////////////////////////////////////////////////////////////////////
    hRet := DirectDrawCreate(nil, g_pDD, nil);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'DirectDrawCreateEx FAILED');
        Result := FALSE;
        exit;
    end;

    // Get exclusive mode
    hRet := g_pDD^^.SetCooperativeLevel^(g_pDD, Wnd, DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'SetCooperativeLevel FAILED');
        Result := FALSE;
        exit;
    end;

    // Set the video mode to 640x480x8
    hRet := g_pDD^^.SetDisplayMode^(g_pDD, 640, 480, 8);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'SetDisplayMode FAILED');
        Result := FALSE;
        exit;
    end;

    // Create the primary surface
    ZeroMemory(@ddsd, SizeOf(ddsd));
    ddsd.dwSize := SizeOf(ddsd);
    ddsd.dwFlags := DDSD_CAPS;
    ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE;
    hRet := g_pDD^^.CreateSurface^(g_pDD, ddsd, g_pDDSPrimary, nil);
    if hRet <> DD_OK then
    begin
        InitFail(Wnd, hRet, 'CreateSurface FAILED');
        Result := FALSE;
        exit;
    end;

    ddsd.dwSize := SizeOf(ddsd);
    ddsd.dwFlags := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwWidth := 640;
    ddsd.dwHeight := 480;
    g_pDD^^.CreateSurface^(g_pDD, ddsd, g_pDDSOne, nil);
    if g_pDDSOne = nil then
    begin
        InitFail(Wnd, hRet, 'CreateSurface FAILED');
        Result := FALSE;
        exit;
    end;

    g_bIsInitialized := TRUE;
    Result := TRUE;
end;

//-----------------------------------------------------------------------------
// Name: CyclePalette()
// Desc:
//-----------------------------------------------------------------------------
procedure CyclePalette;
var
    reg: array[0..14] of Longint;
    k: Longint;
    Temp: array[0..14] of TPaletteEntry;
begin
    // Rotate the whole palette by 15 colors
    Move(pe[30], Temp, SizeOf(TPaletteEntry) * 15);
    Move(pe[45], pe[30], SizeOf(TPaletteEntry) * 210);
    Move(Temp, pe[240], SizeOf(TPaletteEntry) * 15);

    for k := 2 to 16 do
    begin
        reg[k - 2] := pe[15 * k + 14].peRed;
        pe[15 * k + 14].peRed := pe[15 * k + 13].peRed;
        pe[15 * k + 13].peRed := pe[15 * k + 12].peRed;
        pe[15 * k + 12].peRed := pe[15 * k + 11].peRed;
        pe[15 * k + 11].peRed := pe[15 * k + 10].peRed;
        pe[15 * k + 10].peRed := pe[15 * k + 9].peRed;
        pe[15 * k + 9].peRed := pe[15 * k + 8].peRed;
        pe[15 * k + 8].peRed := pe[15 * k + 7].peRed;
        pe[15 * k + 7].peRed := pe[15 * k + 6].peRed;
        pe[15 * k + 6].peRed := pe[15 * k + 5].peRed;
        pe[15 * k + 5].peRed := pe[15 * k + 4].peRed;
        pe[15 * k + 4].peRed := pe[15 * k + 3].peRed;
        pe[15 * k + 3].peRed := pe[15 * k + 2].peRed;
        pe[15 * k + 2].peRed := pe[15 * k + 1].peRed;
        pe[15 * k + 1].peRed := pe[15 * k].peRed;
        pe[15 * k].peRed := reg[k - 2];

        reg[k - 2] := pe[15 * k + 14].peGreen;
        pe[15 * k + 14].peGreen := pe[15 * k + 13].peGreen;
        pe[15 * k + 13].peGreen := pe[15 * k + 12].peGreen;
        pe[15 * k + 12].peGreen := pe[15 * k + 11].peGreen;
        pe[15 * k + 11].peGreen := pe[15 * k + 10].peGreen;
        pe[15 * k + 10].peGreen := pe[15 * k + 9].peGreen;
        pe[15 * k + 9].peGreen := pe[15 * k + 8].peGreen;
        pe[15 * k + 8].peGreen := pe[15 * k + 7].peGreen;
        pe[15 * k + 7].peGreen := pe[15 * k + 6].peGreen;
        pe[15 * k + 6].peGreen := pe[15 * k + 5].peGreen;
        pe[15 * k + 5].peGreen := pe[15 * k + 4].peGreen;
        pe[15 * k + 4].peGreen := pe[15 * k + 3].peGreen;
        pe[15 * k + 3].peGreen := pe[15 * k + 2].peGreen;
        pe[15 * k + 2].peGreen := pe[15 * k + 1].peGreen;
        pe[15 * k + 1].peGreen := pe[15 * k].peGreen;
        pe[15 * k].peGreen := reg[k - 2];

        reg[k - 2] := pe[15 * k + 14].peBlue;
        pe[15 * k + 14].peBlue := pe[15 * k + 13].peBlue;
        pe[15 * k + 13].peBlue := pe[15 * k + 12].peBlue;
        pe[15 * k + 12].peBlue := pe[15 * k + 11].peBlue;
        pe[15 * k + 11].peBlue := pe[15 * k + 10].peBlue;
        pe[15 * k + 10].peBlue := pe[15 * k + 9].peBlue;
        pe[15 * k + 9].peBlue := pe[15 * k + 8].peBlue;
        pe[15 * k + 8].peBlue := pe[15 * k + 7].peBlue;
        pe[15 * k + 7].peBlue := pe[15 * k + 6].peBlue;
        pe[15 * k + 6].peBlue := pe[15 * k + 5].peBlue;
        pe[15 * k + 5].peBlue := pe[15 * k + 4].peBlue;
        pe[15 * k + 4].peBlue := pe[15 * k + 3].peBlue;
        pe[15 * k + 3].peBlue := pe[15 * k + 2].peBlue;
        pe[15 * k + 2].peBlue := pe[15 * k + 1].peBlue;
        pe[15 * k + 1].peBlue := pe[15 * k].peBlue;
        pe[15 * k].peBlue := reg[k - 2];
    end;

    g_pDD^^.WaitForVerticalBlank^(g_pDD, DDWAITVB_BLOCKBEGIN, 0);

    if (g_pDDPal^^.SetEntries^(g_pDDPal, 0, 0, 256, @pe) <> DD_OK) then
        exit;
end;

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Initialization, message loop
//-----------------------------------------------------------------------------
var
  msg: TMsg;
begin
    if not InitApp(hInstance, CmdShow) then
        halt(0);

    ReadBMPIntoSurfaces;
    UpdateFrame;

    while (TRUE) do
    begin
        if PeekMessage(msg, 0, 0, 0, PM_NOREMOVE) then
        begin
            if not GetMessage(msg, 0, 0 ,0) then
                exit;
            TranslateMessage(msg);
            DispatchMessage(msg);
        end else
        if g_bActive then
        begin
            CyclePalette;
        end else
            WaitMessage;
    end;
end.